#!/bin/sh

usage()
{
    echo "
USAGE:
  $(basename $0) [OPTIONS] [OUTPUT.img]

  OPTIONS:
    -b|--board <ax2|ax1|qemu> -- select which output to build
    -B|--boot <boot file to embed>        -- embed boot (sdcard, sw-versions)
    -s|--sign  <key> <cert>               -- produce output.sig signature
    --rootfs <tar archive>                -- rootfs to use (default <OUTPUT>.tar.zst)
    --firmware <image>                    -- add firmware image file (only for main image)
    --installer <image>                   -- produce installer sd card instead of first boot image
    -h|--help

  OUTPUT:
    filename of output. (e.g. alpine.img)
    if signing, also output.sig (e.g. alpine.img.sig)
"
}

error()
{
    echo error: "$@" >&2
    exit 1
}

############
# Veriables
############
signkey=
signcert=
rootfs=
boot=
output=
installer=
board=
extlinux=
firmware=

set -eu

# parse args
while [ $# -ne 0 ]
do
    case "$1" in
    "-b"|"--board")
        [ $# -lt 2 ] && error "$1 requires an argument"
        case "$2" in
        "ax2")
            board=ax2
            arch=aarch64
            suffix=x2
            dtb=armadillo_iotg_g4.dtb
            ;;
        "qemu")
            board=qemu
            arch=x86_64
            suffix=qemu
            ;;
        "ax1")
            board=ax1
            arch=armhf
            suffix=x1
            ;;
        *)
            echo "unknown board($2)"
            usage
            exit 1
            ;;
        esac
        shift
        ;;
    "-B"|"--boot")
        [ $# -lt 2 ] && error "$1 requires an argument"
        boot="$(readlink -e "$2")" \
            || error "boot $2 does not exist"
        shift
        ;;
    "-f"|"--firm"|"--firmware")
        [ $# -lt 2 ] && error "$1 requires an argument"
        firmware="$(readlink -e "$2")" \
            || error "firmware $2 does not exist"
        shift
        ;;
    "-s"|"--sign")
        [ $# -lt 3 ] && error "$1 requires [key cert] arguments"
        signkey="$2"
        signcert="$3"
        [ -r "$signkey" ] || error "key $signkey is not readable"
        [ -r "$signcert" ] || error "cert $signcert is not readable"
        shift 2
        ;;
    "--installer")
        [ $# -lt 2 ] && error "$1 requires an argument"
        installer="$(readlink -e "$2")" \
            || error "install image $2 does not exist"
        [ "$installer" = "${installer%.img}" ] \
            && error "Installer $installer does not end in .img, image type not handled"
        shift
        ;;
    "--rootfs")
        [ $# -lt 2 ] && error "$1 requires an argument"
        rootfs="$(readlink -e "$2")" \
            || error "rootfs $2 does not exist"
        shift
        ;;
    "-h"|"--help")
        usage
        exit 0
        ;;
    -*)
        echo "Invalid option $1" >&2
        usage >&2
        exit 1
        ;;
    *)
        outdir="$(dirname "$(readlink -f "$1")")"
        output="$(basename "$1")"
        [ -z "$rootfs" ] && rootfs="${output%.img}.tar.zst"
        if [ "${output%.img}" = "$output" ]; then
            echo "Output file \"$output\" must end in .img" >&2
            exit 1
        fi
        ;;
    esac
    shift
done

# set default
if [ -z "$board" ]; then
    echo "use default(board=yuakushima-board)"
    board=ax2
    arch=aarch64
    suffix=x2
    dtb=armadillo_iotg_g4.dtb
fi
if [ -z "$output" ]; then
    if [ -z "$rootfs" ]; then
        rootfs=$(ls --sort=time baseos-$suffix-*.tar* 2>/dev/null | grep -vE '\.sig$' | head -n 1)
        # create rootfs if it we didn't find one
        if [ -z "$rootfs" ]; then
            ./build_rootfs.sh -b "$board"
            rootfs=$(ls --sort=time baseos-$suffix-*.tar* 2>/dev/null | grep -vE '\.sig$' | head -n 1)
            [ -n "$rootfs" ] || error "Could not find rootfs that was just built"
        fi
    fi
    outdir="$(pwd)"
    output="${rootfs%.tar*}${installer:+-installer}.img"
    output="$(basename "$output")"
    echo "use default(outdir=$outdir)"
    echo "use default(output=$output)"
fi

[ -e "$rootfs" ] \
    || error "rootfs $rootfs does not exist -- specify with --rootfs or match image name to previously built image"
rootfs=$(realpath -e "$rootfs")

[ "$arch" = "x86_64" ] && extlinux=1
[ -n "$installer" ] && [ -z "$boot" ] && error "--boot must be set for --installer"

# Installs and configures extlinux.
# from https://github.com/alpinelinux/alpine-make-vm-image
setup_extlinux() {
        local mnt="$1"  # path of directory where is root device currently mounted
        local root_dev="$2"  # root device
        local modules="$3"  # modules which should be loaded before pivot_root
        local kernel_flavor="$4"  # name of default kernel to boot
        local serial_port="$5"  # serial port number for serial console
        local default_kernel="$kernel_flavor"
        local kernel_opts=''

        [ -z "$serial_port" ] || kernel_opts="console=$serial_port"

        if [ "$kernel_flavor" = 'virt' ]; then
                _apk search --root . --exact --quiet linux-lts | grep -q . \
                        && default_kernel='lts' \
                        || default_kernel='vanilla'
        fi

        sudo sed -Ei \
                -e "s|^[# ]*(root)=.*|\1=$root_dev|" \
                -e "s|^[# ]*(default_kernel_opts)=.*|\1=\"$kernel_opts\"|" \
                -e "s|^[# ]*(modules)=.*|\1=\"$modules\"|" \
                -e "s|^[# ]*(default)=.*|\1=$default_kernel|" \
                -e "s|^[# ]*(serial_port)=.*|\1=$serial_port|" \
                "$mnt"/etc/update-extlinux.conf

        sudo chroot "$mnt" extlinux --install /boot
        sudo chroot "$mnt" update-extlinux --warn-only 2>&1 \
                | grep -Fv 'extlinux: cannot open device /dev' >&2
}

#######
# Main
#######
cleanup() {
    [ -n "$workdir" ] || return
    [ -d "$workdir/mnt" ] && sudo umount "$workdir/mnt"
    rm -rf "$workdir"
}
trap cleanup EXIT


workdir="$(mktemp -d -t alpine-build-rootfs.XXXXXX)"
cd "$workdir" || error "could not enter workdir"


create_disk() {
    SIZE=300
    [ -n "$installer" ] && SIZE=400
    # x86_64 needs slightly bigger rootfs
    [ "$arch" = "x86_64" ] && SIZE=$((SIZE+50))
    truncate -s $((SIZE+21))M "$output"
    # XXX start 20480 only needed for sd card?
    sgdisk --zap-all --new 1:20480:+$((SIZE))M -c 1:rootfs_0 "$output"
}

create_mount_partition() {
    # extlinux needs the 64bit option to be disabled for some reason
    # (it sometimes works, but is not reliable)
    mkfs.ext4 -F -E offset=$((20480*512)) ${extlinux:+-O "^64bit"} \
        -L rootfs_0 "$output" $((SIZE*1024)) || error "mkfs rootfs"
    mkdir mnt
    sudo mount -o offset=$((20480*512)) "$output" mnt || error "mount rootfs"
}

extract_rootfs() {
    sudo mkdir mnt/boot mnt/mnt mnt/target
    sudo tar -C mnt -xf "$rootfs" || error "extract rootfs"
}


customize_rootfs_first_install() {
    sudo cp -r "$outdir/$board"/first_install/. mnt \
        || error "could not copy first install files"
    if [ -n "$firmware" ]; then
        sudo cp "$firmware" mnt/firm.squashfs \
            || error "Could not copy firmware"
    fi
}

customize_rootfs_installer() {
    local sz sha1

    # kill autoupdate services
    sudo sh -c 'rm -vf mnt/etc/init.d/swupdate* mnt/lib/udev/rules.d/*swupdate*'

    # copy images to install
    sudo cp -r "$outdir/$board"/installer/. mnt || error "could not copy installer files"
    sudo sh -c 'lzop > mnt/image.lzo' < "$installer" || error "Could not copy install image"
    sz=$(stat -c %s "$installer")
    # could use tee to only read image once, but keeping exit stauts is a pain
    sha512=$(sha512sum "$installer")
    sha512=${sha512%% *}
    sudo sh -c "echo $sz $sha512 > mnt/image.sha512sum" || error "Could not write image checksum"
    echo "$(basename "$installer")" \
            | sudo sh -c 'cat > mnt/image.filename' \
        || error "Could not write image filename"
    # round boot up to 4MB
    { cat "$boot"; cat /dev/zero; } \
            | dd bs=1M count=4 iflag=fullblock status=none \
            | sudo sh -c 'lzop > mnt/boot.lzo' \
        || error "Could not write boot image"
    sz=$((4*1024*1024))
    sha512=$({ cat "$boot"; cat /dev/zero; } \
        | dd bs=1M count=4 iflag=fullblock status=none \
        | sha512sum)
    sha512=${sha512%% *}
    sudo sh -c "echo $sz $sha512 > mnt/boot.sha512sum" || error "Could not write boot checksum"
    echo "$(basename "$boot")" \
            | sudo sh -c 'cat > mnt/boot.filename' \
        || error "Could not write boot filename"
}

customize_rootfs() {
    sudo cp -r "$outdir/$board"/installer_common/. mnt \
        || error "could not copy installer common files"
    if [ -n "$installer" ]; then
        customize_rootfs_installer
    else
        customize_rootfs_first_install
    fi
}

install_bootloader() {
    if [ -n "$extlinux" ]; then
        GPTMBR="mnt/usr/share/syslinux/gptmbr.bin"
        [ -e "$GPTMBR" ] || error "extlinux expected but $GPTMBR not found"
        dd bs=440 count=1 conv=notrunc if="$GPTMBR" of="$output"
        sgdisk --attributes=1:set:2 "$output"
        sudo mount --bind /dev mnt/dev
        sudo mount --bind /proc mnt/proc
        setup_extlinux "$PWD/mnt" "LABEL=rootfs_0" "ext4" lts "ttyS0 console=tty1"
        sudo umount mnt/dev
        sudo umount mnt/proc
    fi

    # add link to dtb
    if [ -n "$dtb" ]; then
        if ! [ -e "mnt/boot/$dtb" ]; then
            echo "/boot/$dtb does not exist !!!" >&2
            exit 1
        fi
        sudo ln -s "$dtb" mnt/boot/armadillo.dtb
    fi

    if [ -n "$boot" ]; then
        # write boot for SD card
        dd if="$boot" of="$output" bs=1k seek=32 conv=notrunc
    fi
}

create_disk
create_mount_partition
extract_rootfs
customize_rootfs
install_bootloader

sudo umount mnt || error "could not umount rootfs"
rmdir mnt

# assume VM image for x86_64, in which case create larger sparse file
[ "$arch" = "x86_64" ] && truncate -s 8G "$workdir/$output"

if [ -n "$signkey" ]; then
    openssl cms -sign -in "$workdir/$output" -out "$workdir/$output.sig" \
            -signer "$signcert" -inkey "$signkey" -outform DER \
            -nosmimecap -binary \
        || error "Could not sign image"
fi
[ -e "$workdir/$output.sig" ] && mv "$workdir/$output.sig" "$outdir"
mv "$workdir/$output" "$outdir/"
